/*
*  C Implementation: rfsam
*
* Description: Software RF-Switch for the Amilo M 7400 wireless hardware
*
*
* Author: Marcel Naziri <zwobbl@zwobbl.de>, (c) 2007
*
* Copyright: See COPYING file that comes with this distribution
*
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/param.h>
#include <sys/mman.h>
#include <sys/user.h>
#include <sys/io.h>

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define VERSION  "0.1"

#define BIOS_CODE_ADDR		0x00000000
#define BIOS_CODE_MAPSIZE	0x00100000

#define BIOS_MAGIC_COMMAND	0x9610
#define BIOS_MAGIC_OFF		0x0035
#define BIOS_MAGIC_ON		0x0135

#define PTR_POSITION		5
#define ALLIGNED_STEP		0x10

#define BIOS_SIGN_SIZE		4
static const char unsigned bios_sign[] = { 0x42, 0x21, 0x55, 0x30 };

#define BLOCKSIZE 4096
static unsigned char buffer[BLOCKSIZE];
static size_t buf_len;
static int fd_mem = 0;

void bios_routine_call(int *ax, int *bx, int *cx, void *address) {
    __asm__ __volatile__ (
            "call *%%edi \t\n"
    : "=a"(*ax), "=b"(*bx), "=c"(*cx)
    : "a"(*ax), "b"(*bx), "c"(*cx), "D"(address)
                         );
}

static inline void open_memory()
{
    if ( (fd_mem = open( "/dev/mem", O_RDONLY) ) == -1 ) {
        perror("open_memory");
        exit(EXIT_FAILURE);
    }
}

static inline int get_next_block() 
{
    ssize_t n = read(fd_mem, buffer, BLOCKSIZE);
    if (n == -1) {
        perror("get_next_block");
        exit(EXIT_FAILURE);
    }
    return buf_len = n;
}

static inline int search_bios_signature()
{
    unsigned char *ptr = buffer;
    for ( ; ptr < buffer+buf_len; ptr+=ALLIGNED_STEP)
        if (!memcmp(ptr, bios_sign, BIOS_SIGN_SIZE))
            return ptr - buffer;
    return -1;
}

static void cleanup()
{
    if (fd_mem) close(fd_mem);
}

static inline void usage(char *execname)
{
    fprintf( stderr, "rfsam v" VERSION " - Software RF-Switch for Fujitsu Siemens Amilo M 7400\n");
    fprintf( stderr, "Usage: %s { ON | OFF }\n", execname );
}

int main(int argc, char **argv)
{
    atexit(cleanup); 

    if( argc != 2 ) {
        usage(*argv);
        return EXIT_FAILURE;
    }

    int eax = BIOS_MAGIC_COMMAND;
    int ebx;

    if( !( strcasecmp( argv[1], "ON" ) ) ) {
        ebx = BIOS_MAGIC_ON;
    } else if( !( strcasecmp( argv[1], "OFF" ) ) ) {
        ebx = BIOS_MAGIC_OFF;
    } else {
        usage(*argv);
        return EXIT_FAILURE;
    }

    if (geteuid()) {
        fprintf(stderr, "must be run as root\n");
        return EXIT_FAILURE;
    }

    open_memory();

    int offset = BIOS_CODE_ADDR;
    if (lseek(fd_mem, offset, SEEK_SET) != offset) {
        fprintf(stderr, "seek failed\n");
        exit(EXIT_FAILURE);
    }

    int sign_pos;
    while (get_next_block() && (sign_pos = search_bios_signature()) == -1) {
        offset += BLOCKSIZE;      
        if (offset >= BIOS_CODE_ADDR + BIOS_CODE_MAPSIZE )
            break;
    }

    if (sign_pos == -1) {
        fprintf(stderr, "no bios signature found\n");
        exit(EXIT_FAILURE);
    }

    void *routine_ptr = (void*)buffer + sign_pos + PTR_POSITION;
    offset = *(int*)routine_ptr;

    if (lseek(fd_mem, offset, SEEK_SET) != offset) {
        fprintf(stderr, "seek failed\n");
        exit(EXIT_FAILURE);
    }

    char unsigned bios_routine[8092];
    if (read(fd_mem, bios_routine, 8092) != 8092) {
        fprintf(stderr, "read failed\n");
        exit(EXIT_FAILURE);
    }

    unsigned char *bios_routine_safe = bios_routine + 1;

    iopl(3);

    int dummy = 0;
    bios_routine_call(&eax, &ebx, &dummy,  bios_routine_safe);

    return EXIT_SUCCESS;
}
